/** \file aes_crypt_session.h
  * Autor                 : Bernd Ottmann             01/2012
  */
#pragma once
#ifndef DATAFOX_COMMUNICATION_AES_CRYPT_SESSION_H
#define DATAFOX_COMMUNICATION_AES_CRYPT_SESSION_H

#include "datafox/communication/aes_crypt.h"
#include "datafox/communication/protocol_base.h"
#include <memory.h>

namespace Datafox
{
    namespace Communication
    {
        // Klasse zur Verwaltung einer Session mittels AES.
        class AesCryptSession
        {
        public:
            AesCryptSession()
            {
                // Verschluesselung verwalten
                m_cryptMode = 0;
                m_sessionTime = 0;
                m_sessionTimeout = 0;

                memset( m_sessionKey, 0, sizeof( m_sessionKey ) );
                m_sessionKeySize = 0;

                memset( m_rawCommunicationPassword, 0, sizeof( m_rawCommunicationPassword ) );
                m_rawCommunicationPasswordSize = 0;
                m_globalCommunicationPassword = 1; // Das Passwort, kann global veraendert werden.
            }

        public:
            inline bool isGlobalCommunicationPassword()
            {
                return m_globalCommunicationPassword;
            }
            void resetSessionKey() 
            {
                m_cryptMode = 0;
            }
            // Setzt das Standardpasswort in m_sessionKey
            inline void loadDefaultSessionKey()
            {
                uint8_t sessionKey[16] = {
                    0xc5, 0xc2, 0xc3, 0xc0, 0xc1, 0xd9, 0xdb, 0xf0, 
                    0xcd, 0x07, 0xa7, 0xaf, 0xaa, 0x0a, 0x50, 0xae
                };
                if ( m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE )
                {
                    memcpy( m_sessionKey, m_rawCommunicationPassword, m_rawCommunicationPasswordSize );
                    m_sessionKeySize = m_rawCommunicationPasswordSize;
                }
                else
                {
                    memcpy( m_sessionKey, sessionKey, sizeof( m_sessionKey ) );
                    m_sessionKeySize = sizeof( m_sessionKey );
                }
            }
            inline bool isValidUserKey()
            {
                uint8_t size = m_rawCommunicationPasswordSize;
                if ( size == 0 )
                {
                    // strlen
                    for ( ;size < 16 && m_rawCommunicationPassword[size]; size++ );

                    if ( size == 0 )
                        return false; // Es liegt kein Passwort vor.
                }

                return true;
            }
            void initCryptMode()
            {
                if ( m_cryptMode == IV_PACKET_CRYPT_MODE_NO )
                {
                    // Die verschluesselte Kommunikation wird zwar unterstuetzt,
                    // soll aber nicht genutzt werden.
                }
                else if ( m_cryptMode == IV_PACKET_CRYPT_MODE_DEFAULT || m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE )
                {
                    // Liegt kein Session-Key vor?
                    if ( m_sessionTime == 0 )
                    {
                        loadDefaultSessionKey();
                    }

                    m_aesCryptSend.setKey( m_sessionKey, m_sessionKeySize );
                    m_aesCryptReceive.setKey( m_sessionKey, m_sessionKeySize );
                }
                else
                {
                    // Keine Unterstuetzung der verschluesselten Kommunikation.
                }
            }

            int setCommunicationPassword( const char* password, uint8_t size, uint32_t session_timeout, bool global ) 
            {
                if ( size == 0 )
                {
                    // strlen
                    for ( ;size < 16 && password[size]; size++ );
                }
                else if ( size > 16 )
                {
                    return -IV_ERROR_FUNCTION_PARAMETER;
                }

                int different = 0;
                // Identisches Passwort und auch sessionTimeout sowie global gleich?
                if ( memcmp( m_rawCommunicationPassword, password, size ) != 0 )
                {
                    different++;
                }

                if ( m_rawCommunicationPasswordSize != size )
                {
                    different++;
                }
                if ( m_sessionTimeout != session_timeout )
                {
                    different++;
                }

                if ( m_globalCommunicationPassword != global )
                {
                    // Vorherige Werte sind gleich, dann wird das Flag uebernommen und ok.
                    if ( different == 0 )
                    {
                        m_globalCommunicationPassword = global;
                        return IV_ERROR_SUCCESS;
                    }

                    different++;
                }

                // Liegen unterschiede vor?
                if ( different )
                {
                    // Aktuellen Mode zurücksetzen.
                    resetSessionKey();

                    // Werte uebernehmen
                    memset( m_rawCommunicationPassword, 0, sizeof( m_rawCommunicationPassword ) );
                    memcpy( m_rawCommunicationPassword, password, size );
                    m_rawCommunicationPasswordSize = size;

                    m_sessionTimeout = session_timeout;

                    // Halten ob das Passwort global oder speziell fuer dieses Geraet gesetzt wurde.
                    m_globalCommunicationPassword = global;
                }

                return IV_ERROR_SUCCESS;
            }

        public:
            uint8_t m_cryptMode;
            uint8_t m_rawCommunicationPassword[16];
            uint8_t m_rawCommunicationPasswordSize;
            uint8_t m_sessionKey[16];
            uint8_t m_sessionKeySize;
            uint32_t m_sessionTimeout;
            bool m_globalCommunicationPassword;
            time_t m_sessionTime;

        public:
            uint16_t sendEncrypt( uint8_t *data, uint16_t raw_size )
            {
                return m_aesCryptSend.encrypt( data, raw_size );
            }
            uint32_t sendEncryptStream( uint8_t *data, uint32_t raw_size )
            {
                return m_aesCryptSend.encryptStream( data, raw_size );
            }
            uint8_t sendDecrypt( uint8_t *data, uint16_t *size )
            {
                return m_aesCryptSend.decrypt( data, size );
            }
            uint8_t sendDecryptStream( uint8_t *data, uint32_t *size )
            {
                return m_aesCryptSend.decryptStream( data, size );
            }

            uint16_t receiveEncrypt( uint8_t *data, uint16_t raw_size )
            {
                return m_aesCryptReceive.encrypt( data, raw_size );
            }
            uint32_t receiveEncryptStream( uint8_t *data, uint32_t raw_size )
            {
                return m_aesCryptReceive.encryptStream( data, raw_size );
            }
            uint8_t receiveDecrypt( uint8_t *data, uint16_t *size )
            {
                return m_aesCryptReceive.decrypt( data, size );
            }
            uint8_t receiveDecryptStream( uint8_t *data, uint32_t *size )
            {
                return m_aesCryptReceive.decryptStream( data, size );
            }

        private:
            AesCrypt m_aesCryptSend;
            AesCrypt m_aesCryptReceive;
        };
    } // namespace Communication
} // namespace Datafox
#endif
