/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/module_base.h"
#include <boost/format.hpp>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connection_id Nummer der Verbindungsklasse fuer z.B. Trace.
*/
Datafox::Communication::ModuleBase::ModuleBase( int32_t connectionNumber, uint8_t deviceAddress )
    : m_protocol( std::make_shared<Protocol>( connectionNumber, deviceAddress ) )
    , m_trace( connection().tracePointer() )
#ifdef WITH_VIRTUAL_DEVICE
    , m_uuid( boost::uuids::random_generator()() )
#endif
{
    // Zustand des Kanalobjekts
    m_activeModeState = ams_unused;

    m_result = IV_ERROR_SUCCESS;

    // Callbacks intialisieren
    m_callBackFunctionWithParameters = nullptr;
    m_callBackBlockTransfer = nullptr;
    m_blockTransferCount = 1;
    m_blockTransferNumber = 1;
    protocol().remoteCall = std::bind( &ModuleBase::remoteCall, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4 );

    // Kein Autobaud bei Verbindungsherstellung.
    setAutoBaud( false );
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Datafox::Communication::ModuleBase::~ModuleBase() 
{

}

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Datafox::Communication::ModuleBase::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    m_result = IV_ERROR_SUCCESS;
    protocol().reset();
    std::fill_n( m_packetData, sizeof( m_packetData ), 0 );
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion mit Parametern.
*/
void Datafox::Communication::ModuleBase::setCallBackFunctionWithParameters( std::function<int32_t ( uint32_t size, uint32_t count)> func )
{
    trace( "%s(%i):INFO(3): %s( --omitted-- )\n", __FILE__, __LINE__, __func__ );

    m_callBackFunctionWithParameters = func;
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion mit Parametern.
*/
void Datafox::Communication::ModuleBase::setCallBackBlockTransfer( std::function<int32_t( int32_t connectionNumber, int32_t deviceAddress, uint32_t blockCount, uint32_t blockNumber, uint32_t sizeBlock, uint32_t sizeTransfered )> func )
{
    trace( "%s(%i):INFO(3): %s( --omitted-- )\n", __FILE__, __LINE__, __func__ );

    m_callBackBlockTransfer = func;
}

//#########################################################
//####################### ERROR ###########################
//#########################################################

// ##################################################################
/**
* @brief Die gesamte Fehlerbehandlung zuruecksetzen.
*/
void Datafox::Communication::ModuleBase::resetError()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Fehlernummern loeschen
    m_errorModuleNumber = 0;
    m_errorFunctionNumber = 0;

    // Puffer die in Gebrauch waren, loeschen.
    m_errorModuleText.clear();
    m_errorFunctionText.clear();

    // Reset des Fehlerzustandes der Protokollklasse
    protocol().resetError();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::ModuleBase::errorModule( uint16_t module, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %p )\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Zuweisen eines modulübergreifenden Textes zum Modulfehler.
    switch ( error )
    {
    case 1:
        text = boost::str( boost::format( "( %u ): Das geforderte Modul wird nicht unterstützt." ) % error );
        break;

    case 2:
        text = boost::str( boost::format( "( %u ): Die geforderte Funktion wird nicht unterstützt." ) % error );
        break;

    case 3:
        text = boost::str( boost::format( "( %u ): Es wurden bei der Funktion zu wenige Paramteter angegeben." ) % error );
        break;

    case 4:
        text = boost::str( boost::format( "( %u ): Es wurden bei der Funktion zu viele Paramteter angegeben." ) % error );
        break;

    case 5:
        text = boost::str( boost::format( "( %u ): Geforderte Datenmenge wurde unterschritten." ) % error );
        break;

    case 6:
        text = boost::str( boost::format( "( %u ): Geforderte Datenmenge wurde überschritten." ) % error );
        break;

    case 7:
        text = boost::str( boost::format( "( %u ): Gesamt CRC des Paketes ist falsch.." ) % error );
        break;

    case 8:
        text = boost::str( boost::format( "( %u ): Es werden zusätzliche Steuerinformationen benötigt." ) % error );
        break;

    case 9:
        text = boost::str( boost::format( "( %u ): Die gesendete Protokollversion ist unbekannt." ) % error );
        break;

    case 10:
        text = boost::str( boost::format( "( %u ): Es wurde eine Anfrage erwartet, Antwort erhalten." ) % error );
        break;

    case 11:
        text = boost::str( boost::format( "( %u ): Es wurde eine Antwort erwartet, Anfrage erhalten." ) % error );
        break;

    case 12:
        text = boost::str( boost::format( "( %u ): Paketgroesse und damit die reservierte Puffergroesse ueberschreitet." ) % error );
        break;

    case 13:
        text = boost::str( boost::format( "( %u ): Entschluesselung des Paketinhaltes bringt Fehler." ) % error );
        break;

    case 14:
        text = boost::str( boost::format( "( %u ): Geraet arbeitet verschluesselt und Anfrage war unverschluesselt." ) % error );
        break;

    case 15:
        text = boost::str( boost::format( "( %u ): Bei der Ausführung ist ein interner Timeout eingetreten." ) % error );
        break;

    case 16:
        text = boost::str( boost::format( "( %u ): Angesprochener Slot ist nicht vorhanden.") % error );
        break;

    case 17:
        text = boost::str( boost::format( "( %u ): Kein Timeboy im Slot vorhanden.") % error );
        break;

    case 18:
        text = boost::str( boost::format( "( %u ): Timeboytyp noch nicht erkannt.") % error );
        break;

    case 19:
        text = boost::str( boost::format( "( %u ): Es wird versucht ein TimeboyI-III anzusprechen.") % error );
        break;

    case 20: // Dockingstation liest Datensätze vom Timeboy oder überträgt Listen.
        text = boost::str( boost::format( "( %u ): Ausführung konnte wegen belegtem Kommunikationskanal nicht weitergereicht werden.") % error );
        break;

    case 21:
        text = boost::str( boost::format( "( %u ): Zur Verarbeitung steht nicht genügend Speicher zu Verfügung." ) % error );
        break;

    case 28:
        text = boost::str( boost::format( "( %u ): Das angegebene Passwort stimmt nicht mit dem benötigten überein. Es wird ein Passwort gefordert.") % error );
        break;

    case 29:
        text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
        break;
    }

    if ( text.empty() )
    {
        text = boost::str( boost::format( "( %u ): Nicht abgefangener Modulfehler!" ) % error );
    }

    // Letzte Abfrage halten.
    m_errorModuleNumber = error;
    m_errorModuleText = text;
}


// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::ModuleBase::errorFunction( uint16_t module, uint16_t func, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p )\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    // Sonderfehler für Passwortschutz abhandeln.
    switch ( error )
    {
        case 28:
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %u ): Korrektes Passwort für die Kommunikation erforderlich.") % error );
            break;
        }

        case 29:
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
            break;
        }
    }

    if ( text.empty() )
    {
        text = boost::str( boost::format( "( %u ): Nicht abgefangener Funktionsfehler!" ) % error );
    }

    // Letzte Abfrage halten.
    m_errorFunctionNumber = error;
    m_errorFunctionText = text;
}

// ##################################################################
/**
* @brief Logen der textuellen Fehlerentsprechung.
*
*/
void Datafox::Communication::ModuleBase::traceError( uint8_t device_address, uint16_t module, uint16_t func, int32_t error )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u )\n", __FILE__, __LINE__, __func__, device_address, module, func );

    std::string text;

    if ( error < 0 )
    {
        // Fehlertext ermitteln und eintragen.
        protocol().errorCommunication( -error, text );

        trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, -error & 0xFFFF, text.c_str() );
    }
    else if ( error > 0 )
    {
        // Fehlertext ermitteln und eintragen.
        if ( IV_RESULT_MODULE(error) )
        {
            // Modulefehlertext ermitteln
            errorModule( module, IV_RESULT_MODULE(error), text );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, error & 0xFFFF, text.c_str() );
        }
        else
        {
            // Funktionsfehlertext ermitteln
            errorFunction( module, func, IV_RESULT_FUNCTION(error), text );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, error & 0xFFFF, text.c_str() );
        }
    }
}

//#########################################################
//###################### HARDWARE #########################
//#########################################################

// ##################################################################
/**
* @brief Ausfuehrung einer Modulfunktion in ihrem jeweiligen Kontext.
*
*/
int32_t Datafox::Communication::ModuleBase::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u, %d )\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, contextValue );

    return protocol().call( context, module, func, buffer, count, contextValue );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geräteadresse.
*/
int32_t Datafox::Communication::ModuleBase::remoteCall( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count )
{
    (void)data;
    (void)count;

    if ( module == IV_MODULE_BASE )
    {
        switch ( func )
        {
        case IV_FUNCTION_PING:

        // Funktion nicht definiert
        default:
            return 0x0200;
        }
    }

    // Fehler: Das geforderte Modul wird nicht unterstuetzt.
    return 0x0100;
}

// ##################################################################
/**
* @brief Prueft die Erreichbarkeit des Kommunikationspartners.
*
* @param deviceAddress Anzusprechender Kommunikationspartner.
*/
int32_t Datafox::Communication::ModuleBase::ping( uint8_t device_address, uint8_t *data, uint16_t *count )
{
    trace( "%s(%i):INFO(3): %s( %u, %p, %u )\n", __FILE__, __LINE__, __func__, device_address, data, count );
    int32_t result = IV_ERROR_SUCCESS;
    uint16_t lCount = 0;
    uint32_t auto_baud_rate[5] = { 38400, 9600, 115200, 57600, 19200 };
    uint8_t loop;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        protocol().setDeviceNumberReceiver( device_address );

        // Zu sendende Daten kopieren?
        if ( data && count && *count > 0 )
        {
            lCount = *count;
            memcpy( m_packetData, data, lCount );
        }

        ConnectionBase::ConnectionType connection_type;
        std::string connection_string;
        uint32_t connection_value;
        uint32_t previous_connection_value;
        uint16_t communication_read_timeout;
        uint16_t previous_communication_read_timeout;
        connection().openParameters( connection_type, connection_string, connection_value, communication_read_timeout );
        previous_connection_value = connection_value;
        previous_communication_read_timeout = communication_read_timeout;

        for ( loop = 0; loop < (sizeof( auto_baud_rate ) / sizeof( auto_baud_rate[0] )); ++loop )
        {
            // Bei Autobaud und Seriell wird die Kommunikation in folgender Reihenfolge abgefragt.
            // 1. Voreingestellter Baudwert.
            // Je nach Voreinstellung folgende Prio.
            // - 38400 bd
            // - 9600 bd
            // - 115200 bd
            // - 57600 bd
            // - 19200 bd
            if ( autoBaud() && connection_type == ConnectionBase::ConnectionType::serial )
            {
                // Ab zweitem loop auf naechste Baudrate aus dem Pool wechseln.
                if ( loop > 0 )
                {
                    if ( auto_baud_rate[0] )
                    {
                        connection_value = auto_baud_rate[0];
                    }
                    else if ( auto_baud_rate[1] )
                    {
                        connection_value = auto_baud_rate[1];
                    }
                    else if ( auto_baud_rate[2] )
                    {
                        connection_value = auto_baud_rate[2];
                    }
                    else if ( auto_baud_rate[3] )
                    {
                        connection_value = auto_baud_rate[3];
                    }
                    else if ( auto_baud_rate[4] )
                    {
                        connection_value = auto_baud_rate[4];
                    }
                    else
                    {
                        break; // Abbruch der Schleife, wenn alle zu pruefenden durch sind.
                    }

                    // Schnittstellenbaud anpassen
                    connection().changeSerialState( connection().serialPort(), connection_value, communication_read_timeout );
                }

                // Aktuell gewaehlte Baudrate aus dem Pool nehmen.
                if ( connection_value == auto_baud_rate[0] )
                {
                    auto_baud_rate[0] = 0;
                }
                else if ( connection_value == auto_baud_rate[1] )
                {
                    auto_baud_rate[1] = 0;
                }
                else if ( connection_value == auto_baud_rate[2] ) 
                {
                    auto_baud_rate[2] = 0;
                }
                else if ( connection_value == auto_baud_rate[3] )
                {
                    auto_baud_rate[3] = 0;
                }
                else if ( connection_value == auto_baud_rate[4] )
                {
                    auto_baud_rate[4] = 0;
                }
            }
            else
            {
                // Kein Autobaud gewuenscht oder kein RS232.
                loop = (sizeof( auto_baud_rate ) / sizeof( auto_baud_rate[0] ));
            }

            if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, m_packetData, lCount, 2 )) == IV_ERROR_SUCCESS 
                || (result == -IV_ERROR_NEED_USER_PASSWORD) )
            {
                break;
            }
        }

        // Resultwert hier abpruefen!
        if ( result != IV_ERROR_SUCCESS )
        {
            // Nach Fehlgeschlagener Ermittlung, urspruengliche Einstellung wiederherstellen.
            if ( autoBaud() && connection_type == ConnectionBase::ConnectionType::serial )
            {
                connection().changeSerialState( connection().serialPort(), previous_connection_value, previous_communication_read_timeout );
            }

            // Fehler ausgeben
            traceError( device_address, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, result );
            return result;
        }

        // Empfangene Daten kopieren?
        if ( count )
        {
            *count = lCount;
            if ( data && *count > 0 )
            {
                lCount = *count;
                memcpy( m_packetData, data, lCount );
            }
        }
    }
    else
    {
        // Parameterdaten werden beibehalten

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, data, lCount )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

#ifdef WITH_VIRTUAL_DEVICE
//#########################################################
//############## ID und RESOURCEN #########################
//#########################################################

void Datafox::Communication::ModuleBase::setUuid( boost::uuids::uuid &uuid )
{
    m_uuid = uuid;
}

boost::uuids::uuid Datafox::Communication::ModuleBase::uuid()
{
    return m_uuid;
}

std::string Datafox::Communication::ModuleBase::uuidAsString()
{
    return boost::uuids::to_string( uuid() );
}

void Datafox::Communication::ModuleBase::setPathCache( const boost::filesystem::path &path )
{
    m_pathCache = path;
}

boost::filesystem::path Datafox::Communication::ModuleBase::pathCache()
{
    return m_pathCache;
}

std::string Datafox::Communication::ModuleBase::pathCacheDevice()
{
    boost::filesystem::path path( pathCache() );
    path /= uuidAsString();
    return path.string();
}

std::string Datafox::Communication::ModuleBase::filePathDevice()
{
    boost::filesystem::path path( pathCacheDevice() );
    path /= "device.json";
    return path.string();
}

void Datafox::Communication::ModuleBase::loadFromCache()
{
    // Aktuellen Propertytree lesen.
    try
    {
        if ( boost::filesystem::exists( filePathDevice() ) )
        {
            boost::property_tree::read_json( filePathDevice(), m_deviceProperties );
        }
    }
    catch ( const boost::filesystem::filesystem_error& e )
    {
        trace( "filesystem_error( %s )\n", e.what() );
    }
    catch ( const boost::property_tree::json_parser_error& e )
    {
        trace( "json_parser_error( %s )\n", e.what() );
    }
}

void Datafox::Communication::ModuleBase::saveToCache()
{
    // Aktuellen Propertytree schreiben.
    try
    {
        if ( !boost::filesystem::exists( pathCacheDevice() ) )
        {
            boost::filesystem::create_directories( pathCacheDevice() );
        }

        boost::property_tree::write_json( filePathDevice(), m_deviceProperties );
    }
    catch ( const boost::filesystem::filesystem_error& e )
    {
        trace( "filesystem_error( %s )\n", e.what() );
    }
    catch ( const boost::property_tree::json_parser_error& e )
    {
        trace( "json_parser_error( %s )\n", e.what() );
    }
}
#endif
